#!/usr/bin/env python3

"""
The postgis module contains functions for using PostgreSQL/PostGIS
as data warehouse for data from geonorge.no

Use the module like this:

.. code-block:: python

   import os
   # Import downloader class from module
   from osgeonorge.postgis import PostGISAdapter
   # Initialize class
   ognp = PostGISAdapter(host="my.pghost.no", db_name=my_gis_db",
                                 user=os.getuser, active_schema="reguleringsplaner", cores=5)
   # Import a PostGIS dump from geonorge
   # Define indices to create
   indices = {"omrade": ("omrade", "gist", True),
              "grense": ("grense", "gist", True),
              "posisjon": ("posisjon", "gist", True),
              "arealformal": ("arealformal", "btree", False),
              "kommunenummer": ("kommunenummer", "btree", False),
              "lokalid": ("lokalid", "btree", False),
              "plantype": ("plantype", "btree", False),
              "planstatus": ("planstatus", "btree", False),
              "planidentifiksjon": ("planidentifiksjon", "btree", False),
              "utnyttingstype": ("utnyttingstype", "btree", False),
              "eierform": ("eierform", "btree", False),
              "ikrafttredelsesdato": ("ikrafttredelsesdato", "btree", False),
              "vertikalniva": ("vertikalniva", "btree", False),
              "utnyttingstall": ("utnyttingstall", "btree", False),
              }
   # Import backup file into the active schema
   ognp.import_pg_backup("./Plan_0000_Norge_25833_Reguleringsplaner_PostGIS.backup", indices)
"""


import os
import psycopg2
import re
import subprocess
import warnings
from functools import partial
from multiprocessing import Pool

# Non-builtin imports
from osgeonorge.utils import (
    norwegian_to_ascii,
    compile_ogr_cmd,
    ogr_casts,
    consolidate_attributes,
)


class PostGISAdapter:
    """
    Basic class for importing data from geonorge.no into PostGIS
    Stores, connection parameters, connection and provides methods
    for data import and maintenance

    :todo: a) generalize import of pg_backup
           b) parallelize more
    """

    def __init__(
        self,
        host="localhost",
        port=5432,
        db_name="postgres",
        user=None,
        active_schema="public",
        owner=None,
        reader=None,
        cores=1,
        verbose=True,
    ):
        #: Attribute containg the PostGIS host to connect to
        self.host = host
        #: Attribute defining the port of the PostGIS host to connect to
        self.port = port
        #: Attribute defining the name of the PostGIS database to connect to
        self.db_name = db_name
        #: Attribute defining the name of the user used for connecting to the database
        self.user = user
        #: Attribute containing the name of the existing PostgreSQL user to own the schema and data
        self.owner = owner
        #: Attribute containing the name of the existing PostgreSQL user to have read only priveleges
        self.reader = reader
        #: PostgreSQL connection string used for connections
        self.connection_string = f"host={host} dbname={db_name} user={user}"
        #: Open psycopg2 connection to the specified data base
        self.connection = psycopg2.connect(self.connection_string)
        self.connection.set_session(autocommit=True)
        #: Schema to use for data import
        self.active_schema = active_schema
        #: Attribute containing the number
        self.cores = cores

    def __exit__(self):
        self.connection.close()

    def pg_table_exists(self, table, schema=None):
        """
        Check if a PG table exists

        :param connection_string: Open connection to the PostgreSQL database
        :type connection_string: str
        :param schema: Name of the PostgreSQL schema to look in
        :type schema: str
        :param table: Name of the PostgreSQL table to look for
        :type table: str

        :returns: Verification of table existence
        :rtype: bool
        """

        if not schema:
            schema = self.active_schema

        with self.connection.cursor() as cur:
            cur.execute(
                f"""SELECT EXISTS (SELECT FROM information_schema.tables
                            WHERE table_schema = '{schema}' AND table_name = '{table}');"""
            )
            res = cur.fetchone()[0]
        return res

    def list_pg_backup_content(self, pgbackup_file):
        """
        List content of a PostgreSQL dump from Geonorge
        For parallel import the number of cores to use can be specified

        :param pgbackup_file: Path to a PostgreSQL dump file to list content of
        :type pgbackup_file: str

        :returns: tuple containing content and a list of tables
        :rtype: tuple
        """

        p = subprocess.Popen(
            ["pg_restore", "-l", pgbackup_file], stdout=subprocess.PIPE
        )
        content = p.communicate()[0].decode("UTF8").split("\n")
        table_list = [
            c.split(" ")[4:6] for c in content if " TABLE " in c and " DATA " not in c
        ]
        return content, table_list

    def import_pg_backup(self, pg_backup_file, indices, table_prefix=None, target_schema=None):
        """
        Import a PostgreSQL dump from Geonorge into a target schema
        and index a user defined set of columns
        For parallel import the number of cores to use can be specified

        Assumes credentias stored in a .pgpass file

        :param pg_backup_file: Path to a PostgreSQL dump file to import
        :type pg_backup_file: str or GeonorgeAdapter
        :param indices: Dictionary with indices where keys are column names and
                    values index definitions indices for unmatched columns are ignored
        :type indices: dict
        :param target_schema: Name of the schema to import the backup to
        :type target_schema: str

        :returns: 0 in case of success
        :rtype: int
        """
        if type(pg_backup_file).__name__ == "GeonorgeAdapter":
            pg_backup_files = [
                os.path.join(pg_backup_file.work_dir, list(val[1])[0])
                for val in pg_backup_file.current_zip_content
            ]
        else:
            pg_backup_files = [pg_backup_file]

        if not target_schema:
            target_schema = self.active_schema

        for pg_backup_file in pg_backup_files:
            # Get content of the backup file
            content, table_list = self.list_pg_backup_content(pg_backup_file)

            subprocess.run(
                [
                    "pg_restore",
                    "-O",
                    "-x",
                    "-c",
                    "-h",
                    self.host,
                    "-p",
                    f"{self.port}",
                    "-U",
                    self.user,
                    "-d",
                    self.db_name,
                    "-j",
                    f"{self.cores}",
                    pg_backup_file,
                ]
            )

            for table in table_list:
                source_schema = table[0]
                source_table_name = table[1]
                level = table[0].split("_")[1]
                target_table_name = f"{table_prefix}_{source_table_name}_{level}" if table_prefix else f"{source_table_name}_{level}"
                columns = self.get_column_names(source_table_name)
                table_indices = [
                    indices.get(column) for column in columns if indices.get(column)
                ]
                # Drop target table if it exists
                self.pg_drop_table(target_table_name)
                self.pg_drop_table(source_table_name)
                with self.connection.cursor() as cur:
                    # Move table to different schema
                    cur.execute(
                        f"""ALTER TABLE "{source_schema}"."{source_table_name}" SET SCHEMA "{target_schema}";"""
                    )
                    cur.execute(
                        f"""ALTER TABLE "{target_schema}"."{source_table_name}" RENAME TO "{target_table_name}";"""
                    )
                if table_indices:
                    self.run_maintenance(
                        target_schema,
                        target_table_name,
                        table_indices,
                    )

            for old_schema in {table[0] for table in table_list}:
                with self.connection.cursor() as cur:
                    cur.execute(f"""DROP SCHEMA "{old_schema}";""")
        return 0

    def import_files(
        self,
        table_prefix,
        geonorge_adapter,
        schema=None,
        where_clause=None,
        target_crs=25833,
        remove_prefix=None,
        overwrite=False,
    ):
        """
        Import ogr data from GeonoreAdapter to PostgreSQL/PostGIS in a unified way

        :param table_prefix: Prefix for the name of table to import to
        :type table_prefix: str
        :param where_clause: SQL WHERE-clause without the 'WHERE' keyword to limit import
        :type where_clause: str
        :param target_crs: EPSG code of the target Coordinate Reference System
        :type target_crs: int
        :param schema: Name of the PostgreSQL schema to import into
        :type schema: str
        :param remove_prefix: Prefix to remove from object types (if dataset is to be splitted by objecttype)
        :type remove_prefix: str

        :returns: 0 in case of success
        :rtype: int
        """

        import_dict = geonorge_adapter.current_ogr_files_metadata

        if len(import_dict) < 1:
            warnings.warn("Nothing to import, please check the input")
            return None

        reference_columns = None
        if all(
            [
                layer["object_types"]
                for ogr_file in import_dict.values()
                if ogr_file
                for layer in ogr_file.values()
            ]
        ):
            reference_columns = consolidate_attributes(import_dict)

        include_layer = (
            len({layer for val in import_dict.values() for layer in val}) > 1
        )

        import_tuples = []
        for ogr_file, file_metadata in import_dict.items():
            if self.cores <= 1 or (
                len(import_dict) == 1 and len(import_dict[next(iter(import_dict))]) == 1
            ):
                [
                    import_file(
                        self.connection_string,
                        table_prefix,
                        ogr_file,
                        ogr_layer,
                        layer_content,
                        where_clause=where_clause,
                        reference_columns=reference_columns,
                        target_crs=target_crs,
                        schema=self.active_schema,
                        remove_prefix=remove_prefix,
                        include_layer=include_layer,
                        overwrite=overwrite,
                    )
                    for ogr_layer, layer_content in file_metadata.items()
                ]
            else:
                import_tuples.extend(
                    [
                        (table_prefix, ogr_file, ogr_layer, layer_content)
                        for ogr_layer, layer_content in file_metadata.items()
                    ]
                )
        if import_tuples:
            import_with_kwargs = partial(
                import_file,
                self.connection_string,
                where_clause=where_clause,
                reference_columns=reference_columns,
                target_crs=target_crs,
                schema=self.active_schema,
                remove_prefix=remove_prefix,
                include_layer=include_layer,
                overwrite=overwrite,
            )
            with Pool(self.cores) as p:
                p.starmap(import_with_kwargs, import_tuples)

    def schema_check(
        self,
        comment,
        schema=None,
        owner="postgres",
        users=None,
        reader=None,
        update=True,
    ):
        """
        Create a schema (if needed), adds comments and grants (default) access rights

        :param comment: A descriptive comment for schema content
        :type comment: str
        :param schema: Name of the PostgreSQL schema to create or update
        :type schema: str
        :param owner: Name of the PostgreSQL user to own the schema
        :type owner: str
        :param users: A list of names of existing postgres users that should get access
        :type users: list
        :param users: Name of the PostgreSQL (group-) user with read only access to the schema
        :type users: str
        :param update: A boolean value if access rights and comments of an existing schema should be updated
        :type update: bool

        :returns: Verification of table existence
        :rtype: bool
        """
        if not schema:
            schema = self.active_schema

        schema_check_sql = f"""SELECT exists(SELECT schema_name FROM information_schema.schemata WHERE schema_name = '{schema}');"""
        create_schema_sql = f"""CREATE SCHEMA IF NOT EXISTS "{schema}" AUTHORIZATION "{owner}";
                                COMMENT ON SCHEMA "{schema}" IS '{comment}';
                             """
        if reader:
            create_schema_sql += f"""
-- Make SCHEMA usable for user {reader}
GRANT USAGE ON SCHEMA "{schema}" TO {reader};
GRANT SELECT ON ALL TABLES IN SCHEMA "{schema}" TO {reader};
ALTER DEFAULT PRIVILEGES IN SCHEMA "{schema}" GRANT SELECT ON TABLES TO {reader};"""

        if users:
            for pg_user in users:
                create_schema_sql += f"""GRANT ALL ON SCHEMA "{schema}" TO "{pg_user}" WITH GRANT OPTION;
GRANT ALL ON ALL TABLES IN SCHEMA "{schema}" TO "{pg_user}" WITH GRANT OPTION;
ALTER DEFAULT PRIVILEGES IN SCHEMA "{schema}" GRANT ALL ON TABLES TO "{pg_user}" WITH GRANT OPTION;
"""

        with self.connection.cursor() as cur:
            cur.execute(schema_check_sql)
            res = cur.fetchone()
            if not res[0] or update:
                cur.execute(create_schema_sql)

        return 0

    def get_column_names(self, table, schema=None):
        """
        List column names of a given table

        :param connection: An open connection to the PostgreSQL database
        :type connection: An open psycopg2 connection object
        :param schema: Name of the PostgreSQL schema to look in
        :type schema: str
        :param table: Name of the PostgreSQL table to get columns from
        :type table: str

        :returns: A list of column names, an empty list is returned if the bale is not found
        :rtype: list
        """

        if not schema:
            schema = self.active_schema

        with self.connection.cursor() as cur:
            cur.execute(
                f"""SELECT column_name FROM information_schema.columns
                        WHERE table_schema='{schema}' AND table_name='{table}';"""
            )
            col_names = [row[0] for row in cur.fetchall()]
        return col_names

    def run_maintenance(self, table, indices, schema=None):
        """
        Create indices and vacuum table

        :param connection: An open connection to the PostgreSQL database
        :type connection: An open psycopg2 connection object
        :param schema: Name of the PostgreSQL schema to look in
        :type schema: str
        :param table: Name of the PostgreSQL table to run maintenance on
        :type table: str
        :param indices: A list of tuples with three elements: (column name (str), index type (str), cluster (bool))
        :type indices: list of tuples

        :returns: 0 in case of success
        :rtype: int

        """

        if not schema:
            schema = self.active_schema

        table_columns = self.get_column_names(table, schema=schema)

        with self.connection.cursor() as cur:
            # Create indices on the most often used (queried) fields
            for index in indices:
                if not index[0] in table_columns:
                    warnings.warn(f"Column {index[0]} not found in table {table}, skipping...")
                    continue
                print(f"Creating indices on {index[0]} for table {table} ...")
                cur.execute(
                    f'CREATE INDEX IF NOT EXISTS "{table}_{index[0]}" ON "{schema}"."{table}" USING {index[1]} ("{index[0]}")'
                )
                if index[2]:
                    # Cluster indices because no further input is expected
                    print("Clustering index...")
                    cur.execute(
                        f'ALTER TABLE "{schema}"."{table}" CLUSTER ON "{table}_{index[0]}"'
                    )
            # Vacuum Analyse table for making use of indices
            print(f"Analysing indices of table {table} ...")
            cur.execute(f'VACUUM ANALYZE "{schema}"."{table}";')
        return 0

    def pg_drop_table(self, table, schema=None):
        """
        Create indices and vacuum table

        :param schema: Name of the PostgreSQL schema to look in
        :type schema: str
        :param table: Name of the PostgreSQL table to drop
        :type table: str

        :returns: 0 in case of success
        :rtype: int

        """

        if not schema:
            schema = self.active_schema

        with self.connection.cursor() as cur:
            cur.execute(f"""DROP TABLE IF EXISTS "{schema}"."{table}";""")
        return 0


def import_file(
    connection_string,
    table_prefix,
    ogr_file,
    layer,
    content,
    reference_columns=None,
    where_clause=None,
    target_crs=25833,
    schema=None,
    include_layer=True,
    remove_prefix=None,
    overwrite=False,
    unify_attributes=True,
):
    """
    Import ogr data to PostgreSQL/PostGIS in a unified way

    :param table_prefix: Prefix for the name of table to import to
    :type table_prefix: str
    :param ogr_file: Path to an ogr readable file with GIS data to import
    :type ogr_file: str
    :param layer: Name of the layer(s) for the file to import
    :type layer: str
    :param content: A dictionary describing the content of the layer of a given OGR file
    :type content: dict
    :param file_fields: A dictionary describing the attribute columns of the layer of the input file
    :type file_fields: dict
    :param where_clause: SQL WHERE-clause without the 'WHERE' keyword to limit import
    :type where_clause: str
    :param target_crs: EPSG code of the target Coordinate Reference System
    :type target_crs: int
    :param schema: Name of the PostgreSQL schema to import into
    :type schema: str
    :param remove_prefix: Prefix to remove from object types (if dataset is to be splitted by objecttype)
    :type remove_prefix: str

    :returns: 0 in case of success
    :rtype: int
    """
    print(f"Processing {ogr_file}")

    if not schema:
        schema = "public"

    if include_layer and table_prefix:
        table_prefix = f"{table_prefix}_{layer}"

    if not table_prefix and not layer:
        raise ValueError("Either table prefix or layer must be given")

    # Define how projection should be handled
    if content and content["projection_match"]:
        projection_handling = ["-a_srs", f"EPSG:{target_crs}"]
    else:
        projection_handling = ["-t_srs", f"EPSG:{target_crs}"]

    compile_cmd = partial(
        compile_ogr_cmd,
        connection_string,
        schema,
        projection_handling,
        ogr_file,
        overwrite=overwrite,
    )
    kwargs_list = []
    if content and content["fields"] and content["object_types"]:
        for object_type in content["object_types"]:
            table_suffix = re.sub(
                r"^{}".format(remove_prefix),
                "",
                norwegian_to_ascii(object_type).lower(),
            )
            table_name = f"{table_prefix}_{table_suffix}"
            select_sql = []
            if reference_columns:
                layer_reference_columns = reference_columns[layer][object_type]
            for col_name in layer_reference_columns:
                org_col_name = [
                    col
                    for idx, col in enumerate(content["object_types"][object_type])
                    if col.lower() == col_name
                ]
                if not org_col_name:
                    select_sql.append(
                        f"CAST(NULL AS {ogr_casts[layer_reference_columns[col_name]]}) AS {col_name}"
                    )
                else:
                    select_sql.append(
                        f"CAST({org_col_name[0]} AS {ogr_casts[layer_reference_columns[col_name]]}) AS {col_name}"
                    )
            select_where = f"objekttypenavn = '{object_type}'"
            if where_clause:
                select_where = f"({select_where}) AND ({where_clause})"
            sql_string = (
                f'SELECT {", ".join(select_sql)} from polygons WHERE {select_where}'
            )
            kwargs_list.append(
                (
                    table_name,
                    {
                        "where": None,
                        "layer": None,
                        "sql_string": sql_string,
                    },
                )
            )
    else:
        table_name = table_prefix or layer
        kwargs_list.append(
            (
                table_name,
                {
                    "where": where_clause,
                    "layer": layer,
                    "sql_string": None,
                },
            )
        )

    for kwargs in kwargs_list:
        success = False
        count = 1
        while not success and count < 5:
            cmd = compile_cmd(kwargs[0], **kwargs[1])
            if count >= 2:
                cmd.remove("--config")
                cmd.remove("PG_USE_COPY")
                cmd.remove("YES")
            with subprocess.Popen(
                cmd, stderr=subprocess.PIPE, stdout=subprocess.PIPE
            ) as proc:
                stdout, stderr = proc.communicate()
                success = proc.returncode == 0
            count += 1
        if not success:
            raise ChildProcessError(
                f"Command {' '.join(compile_ogr_cmd(connection_string, schema, projection_handling, ogr_file, kwargs[0], **kwargs[1]))} failed!"
            )
    return 0



#!/usr/bin/env python3

"""
The utils module contains simple helper functions that are used in other submodules
"""

import warnings
from itertools import islice
from zipfile import ZipFile

# Non-builtin imports
from osgeo import ogr

#: Define cast order for SQL data types starting being possible to be casted from left to right
casts_order = ["Integer", "Real", "Double precision", "Date", "DateTime", "String"]

#: Define possible type casts for OGR SQL
ogr_casts = {
    "String": "character(0)",
    "Integer": "integer",
    "Date": "date",
    "Time": "time",
    "DateTime": "timestamp",
    "Real": "float",
    "Double precision": "numeric",
}


def slice_dict(data_dict, number_of_chunks):
    """
    Split a ditct in to a given number of chunks
    e.g. in order to distribute chunks to paralel processing

    :param data_dict: A dict object to slice
    :type data_dict: dict
    :param number_of_chunks: Number of chunks to produce (e.g. = number of cores to use)
    :type number_of_chunks: int

    :returns: A copy of the input string with norwegian non-ascii characters replaced
    :rtype: str
    """

    data_iterator = iter(data_dict)
    chunk_size = (
        int(len(data_dict) // number_of_chunks) + 1
        if len(data_dict) // number_of_chunks > 0
        else 0
    )
    for element in range(0, len(data_dict), chunk_size):
        yield {e: data_dict[e] for e in islice(data_iterator, chunk_size)}


def norwegian_to_ascii(string):
    """
    Replace norwegian non-ascii characters

    :param string: A string from that norwegian non-ascii characters should be replaced
    :type string: str

    :returns: A copy of the input string with norwegian non-ascii characters replaced
    :rtype: str

    :example:
    >>> print(norwegian_to_ascii("Gjærsjømåke"))
    'Gjarsjomake'
    """
    replacement_dict = {
        "Å": "A",
        "å": "a",
        "Ø": "O",
        "ø": "o",
        "Æ": "A",
        "æ": "a",
    }
    ascii_string = ""
    for s in string:
        ascii_string += replacement_dict[s] if s in replacement_dict else s
    return ascii_string


def compile_ogr_cmd(
    connection_string,
    schema,
    projection_handling,
    ogr_file,
    table_name,
    layer=None,
    sql_string=None,
    where=None,
    overwrite=False,
):
    """
    Compile ogr2ogr command to load data to PostGIS

    :param connection_string: A string from that norwegian non-ascii characters should be replaced
    :type connection_string: str
    :param schema: Name of the schema to add to
    :type schema: str
    :param connection_string: A string from that norwegian non-ascii characters should be replaced
    :type connection_string: str
    :param table_name: Name of the table to write to
    :type table_name: str
    :param file: Path to the file to import
    :type file: str
    :param layer: Name of the layer to import (alternative to sql_string option)
    :type layer: str
    :param sql_string: SQL select statement to define import from file (alternative to layer option)
    :type sql_string: str
    :param where: SQL WHERE-clause without the 'WHERE' keyword to limit import
    :type where: str
    :param overwrite: Allow overwriting of existing table
    :type overwrite: bool

    :returns: A copy of the input string with norwegian non-ascii characters replaced
    :rtype: str
    """

    if sql_string and layer:
        raise ValueError("sql_string and layer are mutually excusive")
    if sql_string and where:
        raise ValueError("sql_string and where are mutually excusive")
    if not sql_string and not layer:
        raise ValueError("Either sql_string or layer have to be given")

    # Define create or update mode
    cmd = [
        "/home/NINA.NO/stefan.blumentrath/gdal_inst/bin/ogr2ogr",
        "-lco",
        "GEOMETRY_NAME=geom",
        "-lco",
        "FID=gid",
        "-lco",
        "SPATIAL_INDEX=NONE",
    ]

    if overwrite:
        cmd += ["-overwrite"]
    else:
        cmd += [
            "-update",
            "-append",
        ]

    cmd += [
        "--config",
        "PG_USE_COPY",
        "YES",
        *projection_handling,
        "-f",
        "PostgreSQL",
        "-makevalid",
        "-skipfailure",
        "-nlt",
        "PROMOTE_TO_MULTI",
        "-nln",
        f"{schema}.{table_name}",
    ]

    if sql_string:
        cmd += [
            "-sql",
            sql_string,
        ]

    if where:
        cmd += [
            "-where",
            where,
        ]

    cmd += [f"PG:{connection_string} active_schema={schema}", ogr_file]

    if layer:
        cmd += [layer]

    return cmd


def remove_suffixes(string, suffixes):
    """
    Remove suffix of a string (helper function for Python < 3.9 which contains
    that a similar function

    :param string: String to remove suffix from
    :type string: str
    :param suffixes: Suffix(es) to remove from string
    :type suffixes: str or tuple of str

    :returns: A string with suffix removed
    :rtype: str
    """
    if type(suffixes) == str:
        if string.endswith(suffixes):
            string = string[: -len(suffixes)]
    else:
        for suffix in suffixes:
            if string.endswith(suffix):
                string = string[: -len(suffix)]
    return string


def extract_zip(zip_path, workdir="./"):
    """
    Extract all members of a zip archive to the given working directory

    :param zip_path: Path to a zipfile containing files from geonorge
    :type zip_path: str
    :param workdir: Path to the working directory to which files should be unpacked
    :type workdir: str

    :returns: A list of members of the zip file
    :rtype: list
    """
    with ZipFile(zip_path) as zip_file:
        zip_file.extractall(path=workdir)
        return (zip_path, zip_file.namelist())


def extract_metadata(
    ogr_file,
    target_srs=(32633, 25833, 3045, 5973),
    split_by_objtype=False,
    layer_name=None,
):
    """
    Get metadata from a OGR readable file
    incl. column names and types, SRS match

    :param ogr_file: Path to an OGR readable file
    :type ogr_file: str
    :param target_srs: EPSG codes for target SRS
    :type target_srs: list of int
    :param split_by_objtype: Extract object types from attribute column
    :type split_by_objtype: bool
    :param layer_name: Name of the sosi layer to process
    :type layer_name: str

    :returns: Dict with metadata
    :rtype: dict
    """

    attributes_count = {}

    # Open OGR file
    ogr_ds = ogr.Open(ogr_file)

    if not ogr_ds:
        raise OSError(f"Could not open datset {ogr_file}")

    # List layers
    layers = [
        ogr_ds.GetLayerByIndex(idx).GetName() for idx in range(ogr_ds.GetLayerCount())
    ]

    result_dict = {}

    if layer_name:
        if layer_name not in layers:
            warnings.warn(f"Layer {layer_name} not found in {ogr_file}")
            return result_dict
        layers = [layer_name]

    if not layers:
        warnings.warn(f"No layers found to parse in {ogr_file}")
        return result_dict

    for layer_name in layers:
        # Extract relevant layers
        layer = ogr_ds.GetLayerByName(layer_name)
        # Get SRS and check if it matches the target SRS
        srs = layer.GetSpatialRef()
        if srs:
            projection_match = int(srs.GetAttrValue("AUTHORITY", 1)) in target_srs
        else:
            projection_match = None

        # List fields
        layer_def = layer.GetLayerDefn()
        fields = sorted(
            [
                (
                    layer_def.GetFieldDefn(fid).GetName(),
                    layer_def.GetFieldDefn(fid).GetTypeName(),
                )
                for fid in range(layer_def.GetFieldCount())
            ]
        )
        # field_names = [field[0] for field in fields]
        # field_names_lower = [field[0].lower() for field in fields]
        # List object types
        # Get a list of non-null attribute columns per object type
        # Could be limited to relevant object types
        # layer.SetAttributeFilter(f"objekttypenavn = '{object_type}'")
        if split_by_objtype:
            for feature in layer:
                f_dict = feature.items()
                object_type = f_dict["objekttypenavn"]
                if object_type not in attributes_count:
                    attributes_count[object_type] = {}
                for a in f_dict:
                    if (
                        a not in attributes_count[object_type]
                        or not attributes_count[object_type][a]
                    ):
                        attributes_count[object_type][a] = f_dict[a]

                for object_type in attributes_count:
                    attributes_count[object_type] = {
                        key: val
                        for key, val in attributes_count[object_type].items()
                        if val
                    }
        else:
            attributes_count = None

        result_dict[layer_name] = {
            "object_types": attributes_count,
            "projection_match": projection_match,
            "fields": fields,
        }

    ogr_ds = None
    return result_dict


def consolidate_attributes(file_content_dict):
    """
    Consolidates attributes of SOSI layers unifying data types to smallest common data type

    :param file_content_dict: A dictionary with file content from extract_and_parse_sosi function
    :type file_content_dict: dict

    :returns: A copy of the input dictionary with consolidated attributes
    :rtype: dict

    """
    file_fields = {}
    for sos_file, sos_content in file_content_dict.items():
        for layer, sos_metadata in sos_content.items():
            if layer not in file_fields:
                file_fields[layer] = {}
            for object_type in sos_metadata["object_types"]:
                if object_type not in file_fields:
                    file_fields[layer][object_type] = {}

                for field in sos_metadata["fields"]:
                    if field[0] in sos_metadata["object_types"][object_type] and (
                        field[0].lower() not in file_fields[layer][object_type]
                        or casts_order.index(field[1])
                        > casts_order.index(
                            file_fields[layer][object_type][field[0].lower()]
                        )
                    ):
                        file_fields[layer][object_type][field[0].lower()] = field[1]
    return file_fields



#!/usr/bin/env python3

"""
The geonorge module contains functions and global variables for
interaction with ATOM feeds and download API of geonorge.no

Use the module like this:

.. code-block:: python

   # Import module
   from osgeonorge.geonorge import GeonorgeAdapter
   # Initialize class
   gn_download = GeonorgeAdapter()
   # List available feeds in a pretty table format
   gn_download.print_feeds()

This returns something like this:

.. code-block:: python

   AdminstrativeEnheter GML   30-Oct-2019 05:27 1677 https://nedlasting.geonorge.no/geonorge/ATOM-Feeds/AdminstrativeEn...

Next pick a dataset and format to parse and download:

.. code-block:: python

   # Parse feed of dataset 'Sprstrandsoner' in 'FGDB' format
   gn_download.parse_feed('Sprstrandsoner', 'FGDB')
   # Get a list of datasets that cover the whole country
   gn_download.get_data_list(coverage="Landsdekkende", crs=(3045, 5973, 25833))


Finally, download the currently selected dataset(s):

.. code-block:: python

   gn_download.download_datasets()
   gn_download.extract_datasets()
   gn_download.get_files()
   gn_download.get_metadata()

Now you are ready to e.g. import data to PostgreSQL using the
:obj:`~osgeonorge.postgres.PostGISAdapter`

"""

import base64
import os
import urllib.request
import re
import warnings
import zipfile

from functools import partial
from multiprocessing import Pool
from xml.etree import ElementTree as ET
from html.parser import HTMLParser
from http.client import IncompleteRead
from io import BytesIO

# Non-builtin imports
import requests
from osgeonorge.utils import extract_zip, slice_dict, extract_metadata, remove_suffixes
from osgeonorge.sosi import recode_sosi


class GeonorgeAdapter:
    """
    Basic class for listing, downloading, extracting and parsing  data from
    ATOM feeds in geonorge.no

    :todo: a) add option to store logs
           b) option to only update if needed and also clean in future
           Globbing the download_dir could help
           c) add option to remove zip files
           d) add option to use config file
    """

    #: Attribute containg a dict with supported formats and suffix keys for extracting data from zip archives
    ogr_supported_formats = {
        "CSV": [".csv"],
        "FGDB": [".gdb/gdb"],
        "GEOJSON": [".json"],
        "GML": [".gml"],
        "GPX": [".gpx"],
        "GeoJSON": [".json"],
        "SOSI": [".sos"],
        "Shape": [".shp"],
        "SpatiaLite": [".sqlite"],
        "XYZ": ["xyz"],
    }
    #: Attribute containg a dict with supported formats and suffix keys for extracting data from zip archives
    gdal_supported_formats = {
        "BIN": [".bil"],
        "CSV": [".csv"],
        "DEM": [".dem"],
        "GeoTIFF": [".tif", ".tiff", ".gtif", ".gtiff"],
        "IMG": [".img"],
        "JPEG": [".jpg", ".jpeg"],
        "MrSID": [".mr", ".sid"],
        "SpatiaLite": [".sqlite"],
        "TIFF": [".tif", ".tiff", ".gtif", ".gtiff"],
        "XYZ": [".xyz"],
    }
    #: Attribute containg a dict with other supported formats not readable by GDAL/OGR
    other_supported_formats = {
        "POSTGIS": [".backup"],
        "PostGIS": [".backup"],
    }
    #: Attribute containg a tuple with (currently) unsupported or non-spatial formats
    currently_unsupported_formats = (
        "AI",
        "DAT",
        "DLL",
        "EMF",
        "ENH",
        "JSON",
        "NED",
        "PDF",
        "PPTX",
        "S57",
        "XLSX",
    )
    #: Attribute containg a tuple with (currently) tested formats
    currently_tested_formats = ["FGDB"]
    #: Arrtibute containg a tuple of target EPSG codes (defaults to country wide valid EPSG codes)
    target_crs = (32633, 25833, 3045, 5973)

    def __init__(
        self,
        atom_url="https://nedlasting.geonorge.no/geonorge/ATOM-Feeds",
        download_dir="./",
        work_dir="./",
        credits_file=None,
        cores=1,
        verbose=True,
    ):
        #: Attribute containg the URL to the website listing available ATOM feeds from geonorge.no
        self.atom_url = atom_url
        #: Attribute containg aictionary with parsed list of ATOM feeds
        self.feeds = {}
        self.get_feeds()
        #: Attribute containg a set of all available formats in geonorge feeds
        self.available_formats = {k for val in self.feeds.values() for k in val.keys()}
        #: Attribute containg directory to which data is downloaded or written to
        self.download_dir = download_dir
        #: Attribute containg the working directory where zip files are extracted to
        self.work_dir = work_dir
        #: Attribute containg dictionary with metadata of the current, parsed dataset feed
        self.current_feed_dict = {}
        #: Attribute containg dictionary the format of the current, parsed dataset feed
        self.current_feed_format = None
        #: Attribute containg dictionary with metadata of the current, parsed dataset of a selected format
        self.current_data_dict = {}
        #: Attribute containg a list of dictionaries with members of the zip files downloaded and extracted from the current_data_dict
        self.current_zip_content = None
        #: Attribute containg a list of OGR readable files in the zip files downloaded and extracted from the current_data_dict
        self.current_ogr_files = None
        #: Attribute containg a maximum number for retries for incomplete downloads (may happen if connection is closed prematurely or empty chunk of data is send
        self.max_retries = 10
        #: Attribute containg geonorge credits as tuple (username, password)
        self.user_credits = (None, None)
        self.get_geonorge_credits(credits_file)
        #: Number of cores to use for parallel download (if relevant)
        self.cores = cores
        #: Attribute to define if verbose messages should be given
        self.verbose = verbose
        # Check if download directory is writable
        self.__check_permissions(self.download_dir, "Download")
        # Check if working directory is writable
        self.__check_permissions(self.work_dir, "Extraction of zip files")

    def __check_permissions(self, directory, mode, log_level="warning"):
        """"""
        if not os.access(directory, os.W_OK):
            user_message = "Cannot write to directory {directory}. {mode} will fail."
            if log_level == "warning":
                warnings.warn(user_message)
            else:
                raise OSError(user_message)

    def get_feed(self, dataset, data_format):
        """
        Extract dataset information from ATOM feed and stores the meatadata
        parsed into a dictionary in the
        :attr:`~osgeonorge.geonorge.GeonorgeAdapter.current_feed_dict` attribute.

        Currently parsing is mostly tailored at "kommuneplan" feed

        Basic elements handled so far are:

        - '{http://www.w3.org/2005/Atom}subtitle'
        - '{http://www.w3.org/2005/Atom}link'
        - '{http://www.w3.org/2005/Atom}updated'
        - '{http://www.w3.org/2005/Atom}generator'
        - '{http://www.w3.org/2005/Atom}rights'
        - '{http://www.w3.org/2005/Atom}entry'
        - '{http://www.w3.org/2005/Atom}title'
        - '{http://www.w3.org/2005/Atom}id'

        Datasets are listed as entries

        :param dataset: Dataset of which the ATOM feed XML document should be parsed
        :type dataset: str
        :param data_format: OGR readable format to fetch ATOM feed for
                           (see :attr:`~osgeonorge.geonorge.GeonorgeAdapter.available_formats`).
        :type data_format: str

        """
        if (
            data_format not in self.ogr_supported_formats
            and data_format not in self.other_supported_formats
        ):
            raise ValueError(
                f"Requested format '{data_format}' is currently not supported."
            )
        if data_format not in self.available_formats:
            raise ValueError(
                f"Requested format '{data_format}' is currently not available at geonorge.no."
            )
        if dataset not in self.feeds:
            raise ValueError(f"No ATOM feed found for dataset {dataset} on geonorge.")
        if data_format not in self.feeds[dataset]:
            raise ValueError(
                f"Format {data_format} not found for dataset {dataset} on geonorge."
            )

        feed_dict = {}
        feed = requests.get(self.feeds[dataset][data_format]["url"])
        root = ET.fromstring(feed.text)
        for entry in root.findall("{http://www.w3.org/2005/Atom}entry"):
            updated = entry.find("{http://www.w3.org/2005/Atom}updated").text
            title = [
                e.strip()
                for e in entry.find("{http://www.w3.org/2005/Atom}title").text.split(
                    ","
                )
            ]
            url = entry.find("{http://www.w3.org/2005/Atom}id").text.replace(
                f"_{updated}", ""
            )
            crs = [
                e.attrib["term"]
                for e in entry.findall("{http://www.w3.org/2005/Atom}category")
                if "scheme" in e.attrib
                and e.attrib["scheme"] == "http://www.opengis.net/def/crs/"
            ][0]
            feed_dict[url.split("/")[-1]] = {
                "updated": updated,
                "title": title,
                "url": url,
                "crs": crs,
            }

        self.current_feed_dict = feed_dict
        self.current_feed_format = data_format

    def get_feeds(self):
        """
        Method to fetch available ATOM feeds and store the results as a dict
        in the :attr:`~osgeonorge.geonorge.GeonorgeAdapter.feeds` attribute.

        :param atom_url: URL to ATOM feed XML document to parse
        :type atom_url: str
        """
        atom_list_dict = {}
        parse_atom = ParseATOMList()
        with urllib.request.urlopen(self.atom_url) as response:
            parse_atom.feed(response.read().decode())
        for key, val in parse_atom.feed_dict.items():
            key_elements = key.replace(".xml", "").split("_AtomFeed")
            if len(key_elements) != 2:
                continue
            if key_elements[0] in atom_list_dict:
                atom_list_dict[key_elements[0]][key_elements[1]] = {
                    "url": f"{self.atom_url}/{key}",
                    "updated": val[1][0],
                    "size": val[1][1],
                }
            else:
                atom_list_dict[key_elements[0]] = {
                    key_elements[1]: {
                        "url": f"{self.atom_url}/{key}",
                        "updated": val[1][0],
                        "size": val[1][1],
                    }
                }

        self.feeds = atom_list_dict

    def get_geonorge_credits(self, credits_file=None):
        """
        Method to fetch user credits for Geonorge.no and store it in the
        :attr:`~osgeonorge.geonorge.GeonorgeAdapter.user_credits` attribute.
        Credentials for authentication to geonorge API can be either given
        using a credtis_file (see .geonorge.example) or by defining
        environment variables:

        :envvar:`GEONORGE_USER`
        :envvar:`GEONORGE_PASSWORD`

        The user's :envvar:`HOME` directoy is allways searched for a .geonorge
        credits file.

        Username and password need to represent BAAT authentication (and not GeoID)

        :param credits_file: Path to credis file to read
        :type credits_file: str
        """
        # Get authentication
        user = os.environ.get("GEONORGE_USER")
        password = os.environ.get("GEONORGE_PASSWORD")

        credits_file = credits_file or os.path.expanduser("~/.geonorge")
        if os.path.exists(credits_file):
            try:
                with open(credits_file, "r") as geonorge_credits:
                    user, password = geonorge_credits.read().split("\n")[0:2]
            except OSError as e:
                raise e
        if not all((user, password)):
            warnings.warn(
                "No authentication provided. Downloading data is thus not possible.\n"
                "Please provide authentication information"
            )
        self.user_credits = (user, password)

    def print_feeds(self, pattern=None):
        """
        Method to print available ATOM feeds in table like style

        :param pattern: Regular expression for filtering feeds by dataset name
        :type pattern: str
        """
        for key, val in self.feeds.items():
            for k, v in val.items():
                print(f"{key:<75} {k:<7} {v['updated']:<17} {v['size']:<8} {v['url']}")

    def print_datasets(self):
        """
        Method to print a list of available datasets from ATOM feeds
        """
        print("\n".join(self.feeds.keys()))

    def print_all_formats(self):
        """
        Method to print a list of all available formats from ATOM feeds
        """
        print("\n".join(self.available_formats))

    def print_currect_feed(self):
        """
        Method to print content of an ATOM feed in table like style
        """
        if not self.current_feed_dict:
            warnings.warn("No feed initialized. Run 'get_feed' method first.")
        else:
            for key, val in self.current_feed_dict:
                print(
                    f"{key:<75} {val['updated']:<20} {val['crs'].replace('EPSG:', ''):<5} {val['title'][-1]:<15}"
                )

    def get_data_dict(self, coverage="Landsdekkende", crs=None):
        """
        Method to store metadata of a selected coverage and crs from the
        :attr:`~osgeonorge.geonorge.GeonorgeAdapter.current_feed_dict`
        in the :attr:`~osgeonorge.geonorge.GeonorgeAdapter.current_data_dict`
        attribute. A current feed has to be fetched beforehand with the
        :func:`~osgeonorge.geonorge.GeonorgeAdapter.get_feed` method.

        :param coverage: Coverage of dataset whos feed(s) to parse (allowed values: "Landsdekkende", "Fylkevis", "Kommunevis")
        :type coverage: str
        :param crs: Coordinate reference system of feed(s) to parse
        :type crs: tuple of int
        """

        self.current_data_dict = {
            key: val
            for key, val in self.current_feed_dict.items()
            if (crs is None or val["crs"] in (f"EPSG:{c}" for c in crs))
            and (
                (coverage == "Landsdekkende" and coverage in val["title"][-1])
                or (coverage == "Fylkevis" and val["title"][-1][0].isdigit())
                or (
                    coverage == "Kommunevis"
                    and val["title"][1].isdigit()
                    and "Landsdekkende" not in val["title"][-1]
                )
            )
        }

    def download_current_datasets(self, download=True):
        """
        Method to download data from the
        :attr:`~osgeonorge.geonorge.GeonorgeAdapter.current_data_dict` attribute
        """
        self.__check_permissions(self.download_dir, "Download", log_level="error")
        files_n = len(self.current_data_dict)
        if files_n == 1 or self.cores == 1:
            dict_keys = [
                self._download_geonorge({key: val}, download=download)
                for key, val in self.current_data_dict.items()
            ]
        elif files_n > 1 and self.cores > 1:
            download_with_kwargs = partial(self._download_geonorge, download=download)
            with Pool(self.cores) as p:
                dict_keys = p.map(
                    download_with_kwargs,
                    [
                        dict_chunk
                        for dict_chunk in slice_dict(
                            self.current_data_dict, min(files_n, self.cores)
                        )
                    ],
                )
        else:
            warnings.warn("Nothing to download in current datasets")
        # update dict keys
        if dict_keys:
            dict_keys = [item for tuple_list in dict_keys for item in tuple_list]
            for old_key, new_key in dict_keys:
                self.current_data_dict[new_key] = self.current_data_dict.pop(old_key)

    def _download_geonorge(self, atom_dict, download=True):
        """
        Private method to download all data in the the atom_dict from geonorge nedlasting API

        :todo: Should only update if needed and also clean in future
               Globbing the download_dir could help

        """
        # Setup http session
        s = requests.Session()

        # Check provided authentication
        if (
            len(self.user_credits) != 2
            or type(self.user_credits) != tuple
            or not all(
                [type(auth_element) == str for auth_element in self.user_credits]
            )
        ):
            raise ValueError("No or invalid authentication provided.")

        # Authenticate with basic http authentication by adding header to session
        s.auth = self.user_credits

        dict_key_list = []
        # Download dict content
        for file_id, file_dict in atom_dict.items():
            if self.verbose:
                print(f"Downloading {file_dict['url']}...")
            zip_file = s.get(file_dict["url"], allow_redirects=True, stream=True)
            file_name = zip_file.headers.get("Content-Disposition")
            if file_name and "filename=" in file_name:
                file_name = file_name.split("filename=")[1].strip('"')
            else:
                file_name = file_id
            file_name = os.path.join(self.download_dir, file_name)
            print(file_name)
            if not zip_file.ok:
                zip_file.raise_for_status()
            dict_key_list.append((file_id, file_name))
            if not download:
                continue
            if zip_file.headers["Content-Type"] == "application/zip":
                # Download binary
                with open(file_name, "wb") as f:
                    f.write(zip_file.content)
            elif zip_file.headers["Content-Type"] == "application/octet-stream":
                request = urllib.request.Request(file_dict["url"])
                base64string = base64.b64encode(
                    bytes("%s:%s" % self.user_credits, "ascii")
                )
                request.add_header(
                    "Authorization", "Basic %s" % base64string.decode("utf8")
                )
                retries = 0
                success = False
                data = BytesIO()
                while not success and retries < self.max_retries:
                    try:
                        with urllib.request.urlopen(request) as response:
                            data.write(response.read())
                        success = True
                    except IncompleteRead as e:
                        data.write(e.partial)
                        request.add_header("Range", f"bytes={data.tell()}-")
                        retries += 1

                with open(file_name, "wb") as of:
                    of.write(data.getvalue())

                    """if retries > 0:
                        warnings.warn(
                            f"IncompleteReadError, retry {retries} of max {self.max_retries}"
                        )
                    zip_file = s.get(
                        file_dict["url"],
                        allow_redirects=True,
                        stream=True,
                    )
                    print(zip_file.headers)
                    with open(os.path.join(self.download_dir, file_name), "wb") as f:
                        for content in zip_file.iter_content(1024 * 1024 * 30):
                            f.write(content)
                        f.flush()
                    if float(os.path.getsize()) == float(
                        float(zip_file.headers["Content-Length"])
                    ):
                        success = True
                    else:
                        retries += 1"""
                # with open(os.path.join(self.download_dir, file_name), "wb") as f:
                #     shutil.copyfileobj(zip_file.raw, f)
            else:
                # Download plain text
                with open(file_name, "w") as f:
                    f.write(zip_file.text)
        return dict_key_list

    def extract_current_datasets(self):
        """
        Method to extract data from downloaded zip file(s) from the
        :attr:`~osgeonorge.geonorge.GeonorgeAdapter.current_feed_dict` attribute.
        Datasets have to be downloaded beforehand with the
        :func:`~osgeonorge.geonorge.GeonorgeAdapter.download_datasets` method.
        A list of extracted files is stored in the
        :attr:`~osgeonorge.geonorge.GeonorgeAdapter.current_zip_content` attribute.
        """
        self.__check_permissions(
            self.work_dir, "Extraction of zip files", log_level="error"
        )
        zip_dict = {
            k: v for k, v in self.current_data_dict.items() if zipfile.is_zipfile(k)
        }
        if len(zip_dict) == 1 or self.cores == 1:
            self.current_zip_content = [
                extract_zip(
                    os.path.join(self.download_dir, f"{dataset_id}"),
                    workdir=self.work_dir,
                )
                for dataset_id in zip_dict
            ]
        elif len(zip_dict) > 1 and self.cores > 1:
            with Pool(self.cores) as p:
                self.current_zip_content = p.starmap(
                    extract_zip,
                    [
                        (
                            os.path.join(self.download_dir, f"{dataset_id}"),
                            self.work_dir,
                        )
                        for dataset_id in zip_dict
                    ],
                )
        else:
            warnings.warn("Nothing to extract")

    def get_current_ogr_files(self):
        """
        Method to extract OGR readable file data from downloaded zip files from the
        :attr:`~osgeonorge.geonorge.GeonorgeAdapter.current_feed_dict` attribute.
        Datasets have to be extracted beforehand with the
        :func:`~osgeonorge.geonorge.GeonorgeAdapter.extract_datasets` method.
        """
        ogr_files_in_zips = {}
        for zip_file, zip_content in self.current_zip_content:
            ogr_files_in_zips[zip_file] = []
            for member in zip_content:
                ogr_files_in_zips[zip_file].extend(
                    [
                        os.path.join(
                            self.work_dir, remove_suffixes(member, [f"{os.sep}gdb"])
                        )
                        for suffix in self.ogr_supported_formats[
                            self.current_feed_format
                        ]
                        if member.endswith(suffix)
                    ]
                )

        self.current_ogr_files = ogr_files_in_zips

    def get_current_metadata(self, split_by_objtype=False, layer=None):
        """
        Method to extract metadata from OGR-readable files in the downloaded data

        :attr:`~osgeonorge.geonorge.GeonorgeAdapter.current_ogr_files` attribute.
        Datasets have to be extracted beforehand with the
        :func:`~osgeonorge.geonorge.GeonorgeAdapter.extract_datasets` method
        and ogr readable files listed with the
        :func:`~osgeonorge.geonorge.GeonorgeAdapter.get_ogr_files` method.
        Results are stored in the
        :attr:`~osgeonorge.geonorge.GeonorgeAdapter.current_ogr_files_metadata` attribute.
        """
        ogr_files = []
        for ogr_list in self.current_ogr_files.values():
            ogr_files += ogr_list

        if self.current_feed_format == "SOSI":
            files_n = len(ogr_files)
            if files_n <= 1 or self.cores == 1:
                [recode_sosi(sos_file) for sos_file in ogr_files]
            else:
                with Pool(self.cores) as p:
                    p.map(recode_sosi, ogr_files)

        if len(ogr_files) == 1 or self.cores == 1:
            self.current_ogr_files_metadata = {
                member: extract_metadata(
                    member,
                    target_srs=self.target_crs,
                    split_by_objtype=split_by_objtype,
                    layer_name=layer,
                )
                for member in ogr_files
            }
        else:
            extract_metadata_kwargs = partial(
                extract_metadata,
                target_srs=self.target_crs,
                split_by_objtype=split_by_objtype,
                layer_name=layer,
            )
            with Pool(self.cores) as p:
                file_metadata = p.map(extract_metadata_kwargs, ogr_files)
            self.current_ogr_files_metadata = {
                member: file_metadata[idx] for idx, member in enumerate(ogr_files)
            }

    def fetch(
        self,
        dataset,
        data_format,
        coverage,
        crs=None,
        extract=True,
        get_metadata=True,
        split_by_objtype=False,
        layer=None,
        download=True,
    ):
        """
        Wrapper method to download, extract and parse a dataset of a given
        format with a given coverage. The method assumes the user knows
        and can provide valid input for those three paramters.

        :attr:`~osgeonorge.geonorge.GeonorgeAdapter.current_ogr_files` attribute.
        Datasets have to be extracted beforehand with the
        :func:`~osgeonorge.geonorge.GeonorgeAdapter.extract_datasets` method
        and ogr readable files listed with the
        :func:`~osgeonorge.geonorge.GeonorgeAdapter.get_ogr_files` method.
        Results are stored in the
        :attr:`~osgeonorge.geonorge.GeonorgeAdapter.current_ogr_files_metadata` attribute.

        :param dataset: Name of the dataset to download
        :type dataset: str
        :param dataset: Filter files of a dataset by the given data format
        :type dataset: str
        :param coverage: Filter files of a dataset by the given coverage None means no filter
        :type coverage: str
        :param crs: Filter files of a dataset by the given Coordinate reference systems None means no filter
        :type crs: tuple of int
        """
        self.get_feed(dataset, data_format)
        self.get_data_dict(coverage, crs=crs)
        self.download_current_datasets(download=download)
        if extract:
            self.extract_current_datasets()
        if get_metadata and data_format in self.ogr_supported_formats:
            self.get_current_ogr_files()
            self.get_current_metadata(split_by_objtype=split_by_objtype, layer=layer)


class ParseATOMList(HTMLParser):
    """
    A simple, basic html parser class to extract links to ATOM feeds in geonorge.no
    together with attributes (time of last update, size)
    """

    def __init__(self):
        super().__init__()
        self.reset()
        self.count = 0
        self.tag = None
        self.href = None
        self.feed_dict = {}

    def handle_starttag(self, tag, attrs):
        """
        Extract href tags and stor info a class attribute
        """
        self.tag = tag
        self.count += 1

        if tag == "a":
            self.href = attrs[0][1]
            self.feed_dict[attrs[0][1]] = []

    def handle_data(self, data):
        """
        Get data stored with href tags
        """
        if self.tag == "a":
            self.feed_dict[self.href].append(re.split(r"\s{2,}", data.strip()))



#!/usr/bin/env python3

"""
The postgis module contains functions for using PostgreSQL/PostGIS
as data warehouse for data from geonorge.no

Use the module like this:

.. code-block:: python

   import os
   # Import downloader class from module
   from osgeonorge.postgis import PostGISImporter
   # Initialize class
   pg_importer = PostGISImporter(host="my.pghost.no", db_name=my_gis_db",
                                 user=os.getuser, active_schema="reguleringsplaner", cores=5)
   # Import a PostGIS dump from geonorge
   # Define indices to create
   indices = {"omrade": ("omrade", "gist", True),
              "grense": ("grense", "gist", True),
              "posisjon": ("posisjon", "gist", True),
              "arealformal": ("arealformal", "btree", False),
              "kommunenummer": ("kommunenummer", "btree", False),
              "lokalid": ("lokalid", "btree", False),
              "plantype": ("plantype", "btree", False),
              "planstatus": ("planstatus", "btree", False),
              "planidentifiksjon": ("planidentifiksjon", "btree", False),
              "utnyttingstype": ("utnyttingstype", "btree", False),
              "eierform": ("eierform", "btree", False),
              "ikrafttredelsesdato": ("ikrafttredelsesdato", "btree", False),
              "vertikalniva": ("vertikalniva", "btree", False),
              "utnyttingstall": ("utnyttingstall", "btree", False),
              }
   # Import backup file into the active schema
   pg_importer.import_pg_backup("./Plan_0000_Norge_25833_Reguleringsplaner_PostGIS.backup", indices)
"""


import psycopg2
import subprocess
from osgeonorge.utils import norwegian_to_ascii, compile_ogr_cmd, ogr_casts
import re


class PostGISImporter:
    """
    Basic class for importing data from geonorge.no into PostGIS

    """

    def __init__(
        self,
        host=None,
        port=5432,
        db_name="./",
        user=None,
        active_schema="public",
        owner=None,
        reader=None,
        cores=1,
        verbose=True,
    ):
        #: Attribute containg the PostGIS host to connect to
        self.host = host
        #: Attribute defining the port of the PostGIS host to connect to
        self.port = port
        #: Attribute defining the name of the PostGIS database to connect to
        self.db_name = db_name
        #: Attribute defining the name of the user used for connecting to the database
        self.user = user
        #: Attribute containing the name of the existing PostgreSQL user to own the schema and data
        self.owner = owner
        #: Attribute containing the name of the existing PostgreSQL user to have read only priveleges
        self.owner = owner
        #: PostgreSQL connection string used for connections
        self.connection_string = f"host={host} dbname={db_name} user={user}"
        #: Schema to use for data import
        self.active_schema = active_schema

    def pg_table_exists(self, schema, table):
        """
        Check if a PG table exists

        :param connection_string: Open connectipn to the PostgreSQL database
        :type connection_string: str
        :param schema: Name of the PostgreSQL schema to look in
        :type schema: str
        :param table: Name of the PostgreSQL table to look for
        :type table: str

        :returns: Verification of table existence
        :rtype: bool
        """

        with psycopg2.connect(self.connection_string) as con:
            con.set_session(autocommit=True)
            with con.cursor() as cur:
                cur.execute(
                    f"""SELECT EXISTS (SELECT FROM information_schema.tables
                            WHERE table_schema = '{schema}' AND table_name = '{table}');"""
                )
                res = cur.fetchone()[0]
        return res

    def import_pg_backup(self, pgbackup_file, indices, target_schema):
        """
        Import a PostgreSQL dump from Geonorge into a target schema
        and index a user defined set of columns
        For parallel import the number of cores to use can be specified

        Assumes credentias stored in a .pgpass file

        :param pgbackup_file: Path to a PostgreSQL dump file to import
        :type pgbackup_file: str
        :param indices: Dictionary with indices where keys are column names and
                    values index definitions indices for unmatched columns are ignored
        :type indices: dict
        :param target_schema: Name of the schema to import the backup to
        :type target_schema: str
        :param indices: Number of cores to use for parallel import
        :type indices: int

        :returns: 0 in case of success
        :rtype: int
        """
        p = subprocess.Popen(
            ["pg_restore", "-l", pgbackup_file], stdout=subprocess.PIPE
        )
        content = p.communicate()[0].decode("UTF8").split("\n")
        table_list = [
            c.split(" ")[4:6] for c in content if " TABLE " in c and " DATA " not in c
        ]

        subprocess.run(
            [
                "pg_restore",
                "-O",
                "-x",
                "-c",
                "-h",
                self.host,
                "-p",
                f"{self.port}",
                "-U",
                self.user,
                "-d",
                self.db_name,
                "-j",
                f"{self.cores}",
                pgbackup_file,
            ]
        )

        for table in table_list:
            source_schema = table[0]
            source_table_name = table[1]
            level = table[0].split("_")[1]
            target_table_name = f"{source_table_name}_{level}"
            with psycopg2.connect(self.connection_string) as con:
                con.set_session(autocommit=True)
                columns = get_column_names(con, table[0], table[1])
                table_indices = [
                    indices.get(column) for column in columns if indices.get(column)
                ]
                # Drop target table if it exists
                self.pg_drop_table(con, target_schema, target_table_name)
                with con.cursor() as cur:
                    # Move table to different schema
                    cur.execute(
                        f"""ALTER TABLE "{source_schema}"."{source_table_name}" SET SCHEMA "{target_schema}";"""
                    )
                    cur.execute(
                        f"""ALTER TABLE "{target_schema}"."{source_table_name}" RENAME TO "{target_table_name}";"""
                    )
                if table_indices:
                    run_maintenance(
                        con,
                        target_schema,
                        target_table_name,
                        table_indices,
                    )

        with psycopg2.connect(self.connection_string) as con:
            con.set_session(autocommit=True)
            for old_schema in {table[0] for table in table_list}:
                with con.cursor() as cur:
                    cur.execute(f"""DROP SCHEMA "{old_schema}";""")
        return 0

    def import_file(
        self, schema, table_prefix, ogr_file, content, file_fields, overwrite=False
    ):
        """
        Import ogr data to PostgreSQL/PostGIS in a unified way

        :param ogr_file: Path to an ogr readable file with GIS data to import
        :type ogr_file: str
        :param content: A dictionary describing the file content
        :type content: dict
        :param table_prefix: Prefix for the name of table to import to
        :type table_prefix: str

        :returns: 0 in case of success
        :rtype: int
        """
        print(f"Processing {ogr_file}")

        # Define how projection should be handled
        if content["projection_match"]:
            projection_handling = ["-a_srs", "EPSG:25833"]
        else:
            projection_handling = ["-t_srs", "EPSG:25833"]

        for object_type in content["object_types"]:
            table_suffix = re.sub(r"^kp", "", norwegian_to_ascii(object_type).lower())
            table_name = f"{table_prefix}_{table_suffix}"
            select_sql = []
            reference_columns = file_fields[object_type]
            for col_name in reference_columns:
                org_col_name = [
                    col
                    for idx, col in enumerate(content["object_types"][object_type])
                    if col.lower() == col_name
                ]
                if not org_col_name:
                    select_sql.append(
                        f"CAST(NULL AS {ogr_casts[reference_columns[col_name]]}) AS {col_name}"
                    )
                else:
                    select_sql.append(
                        f"CAST({org_col_name[0]} AS {ogr_casts[reference_columns[col_name]]}) AS {col_name}"
                    )
            select_where = f"objekttypenavn = '{object_type}'"

            sql_string = (
                f'SELECT {", ".join(select_sql)} from polygons WHERE {select_where}'
            )
            success = False
            count = 1
            while not success and count < 5:
                cmd = compile_ogr_cmd(
                    self.connection_string,
                    schema,
                    projection_handling,
                    table_name,
                    ogr_file,
                    sql_string=sql_string,
                    overwrite=overwrite,
                )

                if count >= 2:
                    cmd.remove("--config")
                    cmd.remove("PG_USE_COPY")
                    cmd.remove("YES")
                with subprocess.Popen(
                    cmd, stderr=subprocess.PIPE, stdout=subprocess.PIPE
                ) as proc:
                    stdout, stderr = proc.communicate()
                    success = proc.returncode == 0
                count += 1
            if not success:
                raise ChildProcessError(
                    f"Command {' '.join(compile_ogr_cmd(self.connection_string, schema, projection_handling, sql_string, ogr_file, table_name))} failed!"
                )
        return 0


def schema_check(
    connection_string,
    schema,
    comment,
    owner="postgres",
    users=None,
    reader=None,
    update=True,
):
    """
    Create a schema (if needed), adds comments and grants (default) access rights

    :param connection_string: Open connection to the PostgreSQL database
    :type connection_string: str
    :param schema: Name of the PostgreSQL schema to create or update
    :type schema: str
    :param comment: A descriptive comment for schema content
    :type comment: str
    :param owner: Name of the PostgreSQL user to own the schema
    :type owner: str
    :param users: A list of names of existing postgres users that should get access
    :type users: list
    :param users: Name of the PostgreSQL (group-) user with read only access to the schema
    :type users: str
    :param update: A boolean value if access rights and comments of an existing schema should be updated
    :type update: bool

    :returns: Verification of table existence
    :rtype: bool
    """
    schema_check_sql = f"""SELECT exists(SELECT schema_name FROM information_schema.schemata WHERE schema_name = '{schema}');"""
    create_schema_sql = f"""CREATE SCHEMA IF NOT EXISTS "{schema}" AUTHORIZATION "{owner}";
COMMENT ON SCHEMA "{schema}" IS '{comment}';
"""
    if reader:
        create_schema_sql += f"""
-- Make SCHEMA usable for user {reader}
GRANT USAGE ON SCHEMA "{schema}" TO {reader};
GRANT SELECT ON ALL TABLES IN SCHEMA "{schema}" TO {reader};
ALTER DEFAULT PRIVILEGES IN SCHEMA "{schema}" GRANT SELECT ON TABLES TO {reader};"""

    if users:
        for pg_user in users:
            create_schema_sql += f"""GRANT ALL ON SCHEMA "{schema}" TO "{pg_user}" WITH GRANT OPTION;
GRANT ALL ON ALL TABLES IN SCHEMA "{schema}" TO "{pg_user}" WITH GRANT OPTION;
ALTER DEFAULT PRIVILEGES IN SCHEMA "{schema}" GRANT ALL ON TABLES TO "{pg_user}" WITH GRANT OPTION;
"""
    print(create_schema_sql)
    with psycopg2.connect(connection_string) as con:
        con.set_session(autocommit=True)
        with con.cursor() as cur:
            cur.execute(schema_check_sql)
            res = cur.fetchone()
            if not res[0] or update:
                cur.execute(create_schema_sql)
    return 0


def get_column_names(connection, schema, table):
    """
    List column names of a given table

    :param connection: An open connection to the PostgreSQL database
    :type connection: An open psycopg2 connection object
    :param schema: Name of the PostgreSQL schema to look in
    :type schema: str
    :param table: Name of the PostgreSQL table to get columns from
    :type table: str

    :returns: A list of column names, an empty list is returned if the bale is not found
    :rtype: list
    """
    connection.set_session(autocommit=True)
    with connection.cursor() as cur:
        cur.execute(
            f"""SELECT column_name FROM information_schema.columns
                        WHERE table_schema='{schema}' AND table_name='{table}';"""
        )
        col_names = [row[0] for row in cur.fetchall()]
    return col_names


def run_maintenance(connection, schema, table, indices):
    """
    Create indices and vacuum table

    :param connection: An open connection to the PostgreSQL database
    :type connection: An open psycopg2 connection object
    :param schema: Name of the PostgreSQL schema to look in
    :type schema: str
    :param table: Name of the PostgreSQL table to run maintenance on
    :type table: str
    :param indices: A list of tuples with three elements: (column name (str), index type (str), cluster (bool))
    :type indices: list of tuples

    :returns: 0 in case of success
    :rtype: int

    """
    connection.set_session(autocommit=True)
    with connection.cursor() as cur:
        # Create indices on the most often used (queried) fields
        for index in indices:
            print(f"Creating indices on {index[0]} for table {table} ...")
            cur.execute(
                f'CREATE INDEX IF NOT EXISTS "{table}_{index[0]}" ON "{schema}"."{table}" USING {index[1]} ("{index[0]}")'
            )
            if index[2]:
                # Cluster indices because no further input is expected
                print("Clustering index...")
                cur.execute(
                    f'ALTER TABLE "{schema}"."{table}" CLUSTER ON "{table}_{index[0]}"'
                )
        # Vacuum Analyse table for making use of indices
        print(f"Analysing indices of table {table} ...")
        cur.execute(f'VACUUM ANALYZE "{schema}"."{table}";')
    return 0


def pg_drop_table(connection, schema, table):
    """
    Create indices and vacuum table

    :param connection: An open connection to the PostgreSQL database
    :type connection: An open psycopg2 connection object
    :param schema: Name of the PostgreSQL schema to look in
    :type schema: str
    :param table: Name of the PostgreSQL table to drop
    :type table: str

    :returns: 0 in case of success
    :rtype: int

    """
    connection.set_session(autocommit=True)
    with connection.cursor() as cur:
        cur.execute(f"""DROP TABLE IF EXISTS "{schema}"."{table}";""")
    return 0



#!/usr/bin/env python3

"""
The sosi module contains functions for dealing with data from
geonorge in the Norwegian SOSI format using OSGeo tools
esp. GDAL and the OGR SOSI driver
"""

import os
from zipfile import ZipFile

from osgeo import ogr


def recode_sosi(sosi_path):
    """
    Re-code a SOSI file to encodings supported byt OpenFYBA lib if needed

    :param sosi_path: Path to a SOSI file
    :type sosi_path: str

    :returns: Path to the re-coded SOSI path
    :rtype: str
    """
    sosi_path_iso = sosi_path.replace(".sos", "_iso.sos")
    # Valid encodings according to SOSI standard
    sosi_encodings = (
        "UTF-8",
        "ISO8859-1",
        "ISO8859-10",
        "ANSI",
        "DOSN8",
        "DECN7",
        "ND7",
    )

    # Most comprehensive encoding supported by the GDAL SOSI driver
    target_encoding = "ISO8859-10"
    source_encoding = None
    for encoding in sosi_encodings:
        try:
            with open(sosi_path, "r", encoding=encoding) as sos:
                sosi_content = sos.read()
                source_encoding = encoding
                break
        except UnicodeDecodeError as e:
            print(f"Encoding of {sosi_path} is not {encoding}")
            pass
    if source_encoding is None:
        raise Exception(f"Could not read SOSI file {sosi_path} with valid encodings")
    if source_encoding == "UTF-8":
        # Remove BOM
        if sosi_content.startswith("\ufeff"):
            sosi_content = sosi_content.lstrip("\ufeff")
        sosi_content = sosi_content.split("\n")
        for idx in range(10):
            if "..TEGNSETT " in sosi_content[idx]:
                sosi_content[idx] = f"..TEGNSETT {target_encoding}"
        try:
            with open(sosi_path_iso, "w", encoding=target_encoding) as iso_sos:
                iso_sos.write("\n".join(sosi_content))
        except UnicodeDecodeError as e:
            print(f"Could not encode data in {sosi_file} with {target_encoding}")
            print(f"Got the following error:\n{e}")
            print("Replacing invalid characters")
            with open(
                sosi_path_iso, "w", encoding=target_encoding, errors="replace"
            ) as iso_sos:
                iso_sos.write("\n".join(sosi_content))
        os.replace(sosi_path_iso, sosi_path)
    return sosi_path


def extract_metadata(opath, target_srs=(32633, 25833, 3045), layer_name="polygons"):
    """
    Get metadata from a SOSI file spit by objecttype
    incl. column names and types, SRS match

    :param opath: Path to a SOSI file
    :type opath: str
    :param target_srs: EPSG codes for target SRS
    :type target_srs: list of int
    :param layer_name: Name of the sosi layer to process
    :type layer_name: str

    :returns: Dict with metadata
    :rtype: dict
    """

    attributes_count = {}

    # Open SOSI file
    ogr_ds = ogr.GetDriverByName("SOSI").Open(opath)
    # Get SRS and check if it matches the target SRS
    srs = ogr_ds.GetLayerByIndex(0).GetSpatialRef()
    # List layers
    layers = [
        ogr_ds.GetLayerByIndex(idx).GetName() for idx in range(ogr_ds.GetLayerCount())
    ]
    # Extract relevant layers
    layer = ogr_ds.GetLayerByName(layer_name)
    # List fields
    layer_def = layer.GetLayerDefn()
    fields = sorted(
        [
            (
                layer_def.GetFieldDefn(fid).GetName(),
                layer_def.GetFieldDefn(fid).GetTypeName(),
            )
            for fid in range(layer_def.GetFieldCount())
        ]
    )
    field_names = [field[0] for field in fields]
    field_names_lower = [field[0].lower() for field in fields]
    # List object types
    # Get a list of non-null attribute columns per object type
    # Could be limited to relevant object types
    # layer.SetAttributeFilter(f"objekttypenavn = '{object_type}'")
    for feature in layer:
        f_dict = feature.items()
        object_type = f_dict["objekttypenavn"]
        if object_type not in attributes_count:
            attributes_count[object_type] = {}
        for a in f_dict:
            if (
                a not in attributes_count[object_type]
                or not attributes_count[object_type][a]
            ):
                attributes_count[object_type][a] = f_dict[a]

    ogr_ds = None
    for object_type in attributes_count:
        attributes_count[object_type] = {
            key: val for key, val in attributes_count[object_type].items() if val
        }

    return {
        "object_types": attributes_count,
        "projection_match": srs.GetAttrValue("AUTHORITY", 1) in target_srs,
        "fields": fields,
    }


def extract_and_parse_sosi(zip_file, workdir="./"):
    """
    Extract SOSI files from ZIP, recode if needed and compile metadata

    :param zip_file: Path to a zipfile containing SOSI files
    :type zip_file: str
    :param workdir: Path to the working directory to which files should be unpacked
    :type workdir: str

    :returns: A dictionary with sosi path as ky and etadata as content
    :rtype: dict
    """
    zipname = zip_file
    file_content = {}
    with ZipFile(zip_file) as zip_file:
        members = [
            sos_name for sos_name in zip_file.namelist() if sos_name.endswith(".sos")
        ]
        if len(members) == 0:
            print(f"No data to extract from {zip_file.filename}")
            zip_file.close()
            return

        for member in members:
            # if any([kid in member for kid in ["1576", "1573"]]):
            #     print("Skipping: ", os.path.join(workdir, member))
            #     zip_file.close()
            #     continue
            print("Processing", os.path.join(workdir, member))
            sosi_path = os.path.join(workdir, member)
            if os.path.exists(sosi_path):
                print(
                    f"""File "{os.path.join(workdir, member)} from {zipname} exists."""
                )
            # Extract SOSI file
            zip_file.extract(member, path=workdir)

            # print(f"""Processing file "{os.path.join(workdir, member)} from {zipname}""")

            # Recode SOSI file if required
            opath = recode_sosi(sosi_path)
            # Add file info to metadata dict
            file_content[opath] = extract_metadata(opath)
        zip_file.close()
    return file_content



#!/usr/bin/env python3

"""
The grassgis module contains functions for using GRASS GIS
as data warehouse for data from geonorge.no

This module is incomplete!

Use the module like this:

.. code-block:: python

   import os
   # Import downloader class from module
   from osgeonorge.grassgis import GRASSGISAdapter
   # Initialize class
   ogng = GRASSGISAdapter(grassdata="/data/grassdata", cores=5)

"""


import subprocess
import re
from osgeonorge.utils import norwegian_to_ascii, compile_ogr_cmd, ogr_casts

from grass_session import Session
import grass.script as gscript


class GRASSGISAdapter:
    """
    Basic class for importing data from geonorge.no into GRASS GIS

    """

    def __init__(
        self,
        grassdata="./",
        cores=1,
        verbose=True,
    ):
        #: Attribute containg the path to the GRASS GIS database
        self.grassdata = grassdata
        #: Attribute containg the path to the GRASS GIS database
        self.cores = cores
        #: Attribute containg the name of the current GRASS GIS location
        self.current_location = None
        #: Attribute containg the name of the current GRASS GIS mapset
        self.current_mapset = None

    def import_raster(self, raster_file, raster_map):
        """Method to import raster data into GRASS GIS"""
        with Session(
            gisdb=self.grassdata,
            location=self.current_location,
            mapset=self.current_mapset,
        ):
            gscript.run_command("r.in.gdal", input=raster_file, output=raster_map)

    def import_vector(self, vector_file, vector_map):
        """Method to import vector data into GRASS GIS"""
        with Session(
            gisdb=self.grassdata,
            location=self.current_location,
            mapset=self.current_mapset,
        ):
            gscript.run_command("v.in.ogr", input=vector_file, output=vector_map)



import argparse
import sys

from osgeonorge.geonorge import atom_url, list_feeds

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-u', '--url', type=str, dest="url", default=atom_url, help='URL to list with ATOM feeds')
    args = parser.parse_args()

    list_feeds(args.url)


if __name__ == '__main__':
    main()
